from libc.stdint cimport int64_t


cdef extern from "stdlib.h":
    void free(void *)
    void *malloc(size_t)
    void *calloc(size_t,size_t)
    void *realloc(void *,size_t)
    int c_abs "abs" (int)
    void qsort(void *base, size_t nmemb, size_t size,
            int (*compar)(void *,void *))

cdef extern from "stdio.h":
    ctypedef struct FILE:
        pass

    FILE *fopen(char *,char *)
    FILE *open_memstream(void *, size_t *)
    FILE *fmemopen (void *, size_t, const char *)

    int fclose(FILE *)
    int fflush(FILE *)
    int sscanf(char *str,char *fmt,...)
    int sprintf(char *str,char *fmt,...)
    int fprintf(FILE *ifile,char *fmt,...)
    char *fgets(char *str,int size,FILE *ifile)


cdef extern from "ncls/src/fintervaldb.h":
    ctypedef struct IntervalMap:
        double start
        double end
        int target_id
        int sublist

    ctypedef struct IntervalIterator:
        pass

    ctypedef struct SublistHeader:
        int start
        int len

    int find_overlap_start(double start, double end, IntervalMap im[], int n)
    int imstart_qsort_cmp(void *void_a,void *void_b)
    # int target_qsort_cmp(void *void_a,void *void_b)
    IntervalMap *read_intervals(int n,FILE *ifile)
    SublistHeader *build_nested_list(IntervalMap im[],int n,int *p_n,int *p_nlists)
    SublistHeader *build_nested_list_inplace(IntervalMap im[],int n,int *p_n,int *p_nlists)
    IntervalMap *interval_map_alloc(int n)
    IntervalIterator *interval_iterator_alloc()
    int free_interval_iterator(IntervalIterator *it)
    IntervalIterator *reset_interval_iterator(IntervalIterator *it)
    int *alloc_array(int n)
    int find_intervals_stack(int start_stack[], int end_stack[], int sp, int start,
                             int end, IntervalMap im[], int n,
                             SublistHeader subheader[], IntervalMap buf[],
                             int *nfound)

    int find_intervals(IntervalIterator *it0,
                       double start,
                       double end,
                       IntervalMap im[],
                       int n,
                       SublistHeader subheader[],
                       int nlists,
                       IntervalMap buf[],
                       int nbuf,
                       int *p_nreturn,
                       IntervalIterator **it_return)
    void find_k_next(int start, int end,
                    IntervalMap im[], int n,
                    SublistHeader subheader[], int nlists,
                    IntervalMap buf[], int ktofind,
                    int *p_nreturn)



cimport ncls.src.cfncls as cn

cimport cython

from libc.stdint cimport int64_t

cimport ncls.src.cncls as cn

from libc.stdlib cimport malloc
import numpy as np

cdef inline int int_max(int a, int b): return a if a >= b else b
cdef inline int int_min(int a, int b): return a if a <= b else b
# import ctypes as c

try:
    dummy = profile
except:
    profile = lambda x: x


cdef class FNCLS:

    cdef cn.SublistHeader *subheader
    cdef cn.IntervalMap *im
    cdef int n, ntop
    cdef int nlists

    # build NCLS from array of starts, ends, values
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    def __cinit__(self, const double [::1] starts=None, const double [::1] ends=None, const int64_t [::1] ids=None):

        if None in (starts, ends, ids):
            return

        if len(starts) == 0 or len(ends) == 0 or len(ids) == 0:
            return

        cdef int i
        cdef length = len(starts)
        self.close() # DUMP OUR EXISTING MEMORY
        self.n = len(starts)
        self.im = cn.interval_map_alloc(self.n)
        if self.im == NULL:
            raise MemoryError('unable to allocate IntervalMap[%d]' % self.n)
        i = 0
        for i in range(length):
            self.im[i].start = starts[i]
            self.im[i].end = ends[i]
            self.im[i].target_id = <long> ids[i]
            self.im[i].sublist = -1

        self.subheader = cn.build_nested_list(self.im, self.n, &(self.ntop), &(self.nlists))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cpdef all_overlaps_both(self, const double [::1] starts, const double [::1] ends, const int64_t [::1] indexes):

        cdef int i = 0
        cdef int nhit = 0
        cdef int length = len(starts)
        cdef int loop_counter = 0
        cdef int nfound = 0

        output_arr = np.zeros(length, dtype=np.double)
        output_arr_other = np.zeros(length, dtype=np.double)
        cdef double [::1] output
        cdef double [::1] output_other

        output = output_arr
        output_other = output_arr_other

        cdef cn.IntervalIterator *it
        cdef cn.IntervalIterator *it_alloc

        cdef cn.IntervalMap im_buf[1024]
        if not self.im: # if empty
            return [], []

        it_alloc = cn.interval_iterator_alloc()
        it = it_alloc
        for loop_counter in range(length):

            # print("---search---: {} {}".format(starts[loop_counter], ends[loop_counter]))
            while it:
                i = 0
                cn.find_intervals(it, starts[loop_counter], ends[loop_counter], self.im, self.ntop,
                                self.subheader, self.nlists, im_buf, 1024,
                                &(nhit), &(it)) # GET NEXT BUFFER CHUNK

                # print("nhit", nhit)
                if nfound + nhit >= length:

                    length = (length + nhit) * 2
                    output_arr = np.resize(output_arr, length)
                    output_arr_other = np.resize(output_arr_other, length)
                    output = output_arr
                    output_other = output_arr_other

                while i < nhit:
                    # print("  i", i)

                    # print("length", length)
                    # print("nfound", nfound)
                    # print("loop_counter", loop_counter)
                    output[nfound] = indexes[loop_counter]
                    output_other[nfound] = im_buf[i].target_id

                    # print("  hit {} {}".format(im_buf[i].start, im_buf[i].end))
                    # print("  output[nfound]", output[nfound])
                    # print("  output_other[nfound]", output_other[nfound])

                    nfound += 1
                    i += 1

            cn.reset_interval_iterator(it_alloc)
            it = it_alloc

        cn.free_interval_iterator(it_alloc)
        # end = time()

        # print("ncls time:", end - start)

        return output_arr[:nfound], output_arr_other[:nfound]


    def __len__(self):
        return self.n

    def __str__(self):

        contents = ["Number intervals:", self.n, "Number of intervals in main list:", self.ntop, "Number of intervals with subintervals:", self.nlists, "Percentage in top-level interval", self.ntop/float(self.n)]
        return "NCLS64\n------\n" + "\n".join(str(c) for c in contents)

    def __dealloc__(self):
        'remember: dealloc cannot call other methods!'
        if self.subheader:
            cn.free(self.subheader)
        if self.im:
            cn.free(self.im)

    def close(self):
        if self.subheader:
            cn.free(self.subheader)
        if self.im:
            cn.free(self.im)
            self.subheader = NULL
            self.im = NULL

        return None



from setuptools import find_packages, Extension, Command
from distutils.core import setup


# try:
#     from Cython.Build import cythonize
# except ImportError:
#      def cythonize(*args, **kwargs):
#          from Cython.Build import cythonize
#          return cythonize(*args, **kwargs)





import os
import sys

CLASSIFIERS = """Development Status :: 5 - Production/Stable
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows :: Windows NT/2000
Operating System :: OS Independent
Operating System :: POSIX
Operating System :: POSIX :: Linux
Operating System :: Unix
Programming Language :: Python
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Bio-Informatics"""

# split into lines and filter empty ones
CLASSIFIERS = CLASSIFIERS.split("\n")

# macros = [("CYTHON_TRACE", "1")]

# # extension sources
# macros = []

# if macros:
#     from Cython.Compiler.Options import get_directive_defaults
#     directive_defaults = get_directive_defaults()
#     directive_defaults['linetrace'] = True
#     directive_defaults['binding'] = True

dir_path = os.path.dirname(os.path.realpath(__file__))

include_dirs = [dir_path + "/ncls/src", dir_path]

__version__ = open("ncls/version.py").readline().split(" = ")[1].replace(
    '"', '').strip()



extensions = [
    Extension(
        "ncls.src.ncls", ["ncls/src/ncls.pyx", "ncls/src/intervaldb.c"],
        # define_macros=macros,
        include_dirs=include_dirs),
    Extension(
        "ncls.src.ncls32", ["ncls/src/ncls32.pyx", "ncls/src/intervaldb32.c"],
        # define_macros=macros,
        include_dirs=include_dirs),
    Extension(
        "ncls.src.fncls", ["ncls/src/fncls.pyx", "ncls/src/fintervaldb.c"],
        # define_macros=macros,
        include_dirs=include_dirs)]

# using setuptools to cythonize if cython not found
# not recommended by cython docs, but still
try:
    from cython.Build import cythonize
    ext_modules = cythonize(extensions, language_level=2)
except ImportError:
    ext_modules = extensions


setup(
    name = "ncls",
    version=__version__,
    packages=find_packages(),
    ext_modules = ext_modules,
    setup_requires = ["cython"],
    install_requires = ["numpy"],
    # py_modules=["pyncls"],
    description = \
    'A wrapper for the nested containment list data structure.',
    long_description = __doc__,
    # I am the maintainer; the datastructure was invented by
    # Alexander V. Alekseyenko and Christopher J. Lee.
    author = "Endre Bakken Stovner",
    author_email='endrebak85@gmail.com',
    url = 'https://github.com/endrebak/ncls',
    license = 'New BSD License',
    classifiers = CLASSIFIERS,
    package_data={'': ['*.pyx', '*.pxd', '*.h', '*.c']},
    include_dirs=["."],
)
